/*
	Copyright (c) 2004 Pablo Bleyer Kocik.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	1. Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

	2. Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

	3. The name of the author may not be used to endorse or promote products
	derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
	MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
	EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
	IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
*/

import java.util.Vector;
import java.util.TreeSet;
import java.util.ListIterator;
import java.util.Iterator;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

public class Assembler {

	public static final int
		program_size = 1<<10;

	public Vector expressions;
	public int[] program;
	public String[] listing;
	public Environment environment;

	String formatHex(int n, int p) {
		String h = Integer.toHexString(n);
		int hl = p-h.length();
		for (int m=0; m<hl; ++m) h = "0" + h;
		return h;
	}

	Assembler() {
		expressions = null;
		listing = null;
		environment = new Environment();
		program = new int[program_size];
		for (int i=0; i<program.length; ++i) program[i] = 0;
	}

	Assembler(Vector l) {
		environment = new Environment();
		program = new int[program_size];
		for (int i=0; i<program.length; ++i) program[i] = 0;
		expressions = l;
		listing = null;
		parse();
		assemble();
	}

	/** Parse expressions for symbols. */
	void parse() {
		if (expressions == null) return;
		int pc = 0; // program counter
		ListIterator i = expressions.listIterator();
		Vector lbl = null;

		while (i.hasNext()) {
			AsmExpression e = (AsmExpression)i.next();
			if (e.empty()) continue;

			// add label
			if (e instanceof Label) {
				if (lbl == null) lbl = new Vector();
				lbl.add(e);
				// environment.symbols.put(c.name, new Symbol(Symbol.LABEL, pc));
			}
			else if (e instanceof Command) {
				Command c = (Command)e;
				Object current = null;

				// add constants
				if (c.name.equals("constant")) {
					if (c.list == null
						|| c.list.size() != 2
						|| !(c.list.get(0) instanceof String)
						|| !(c.list.get(1) instanceof Integer)
						) throw new IllegalArgumentException();
					current = c.list.get(1);
					environment.symbols.put(
						c.list.get(0), new Symbol(Symbol.CONSTANT, current)
						);
				}
				else if (c.name.equals("address")) {
					if (c.list == null
						|| c.list.size() != 1
						|| !(c.list.get(0) instanceof Integer)
						) throw new IllegalArgumentException();
					current = c.list.get(0);
					pc = ((Integer)current).intValue();
				}
				else if (c.name.equals("register")) {
					if (c.list == null
						|| c.list.size() != 2
						|| !(c.list.get(0) instanceof String)
						|| !(c.list.get(1) instanceof Integer)
						) throw new IllegalArgumentException("Bad register declaration");
					current = c.list.get(1);
					environment.symbols.put(
						c.list.get(0), new Symbol(Symbol.REGISTER, current)
						);
				}
				else {
					current = new Integer(pc);
					++pc;
				}

				// write labels
				if (lbl != null) {
					ListIterator j = lbl.listIterator();
					while (j.hasNext()) {
						Label t = (Label)j.next();
						environment.symbols.put(
							t.name, new Symbol(Symbol.LABEL, current)
							);
					}
					lbl = null;
				}

			}
		}
	}

	/** Assemble a vector of expressions. */
	void assemble() {
		if (expressions == null) return;
		int pc = 0; // program counter
		ListIterator i = expressions.listIterator();
		Vector lst = new Vector();
		int cl = -1;
		Instruction1 ins = new Instruction1();

		while (i.hasNext()) {
			AsmExpression e = (AsmExpression)i.next();
			if (e.empty()) continue;

			if (e instanceof Label) {
				lst.add("// #" + e.line + ": " + e.toString());
				cl = e.line;
			}
			else if (e instanceof Comment) {
				if (e.line == cl) {
					String s = lst.lastElement() + " " + e.toString();
					lst.setElementAt(s, lst.size()-1);
				}
				else lst.add("// #" + e.line + ": " + e.toString());
			}
			else if (e instanceof Command) {
				Command c = (Command)e;
				String s = "";
				cl = c.line;

				if (c.name.equals("address")) {
					pc = ((Integer)c.list.get(0)).intValue();
					lst.add(
						"@" + formatHex(pc,2) + " // #" + c.line + ": " + c.toString()
						);
				}
				else if (!c.name.equals("constant")
					&& !c.name.equals("register")
					) {
					int t;
					try {
						t = ins.toCode(c.name, c.list, environment);
					}
					catch (IllegalArgumentException x) {
						throw new IllegalArgumentException(
							x.getMessage() + " at line " + c.line
							);
					}
					program[pc] = t;
					s += formatHex(program[pc], 4) + " ";
					++pc;
					lst.add(s + "// #" + c.line + ": " + c.toString());
				} else lst.add(s + "// #" + c.line + ": " + c.toString());
			}
		}

		listing = new String[lst.size()];
		for (int j=0; j<lst.size(); ++j) listing[j] = (String)lst.get(j);
	}

	public String toString() {
		StringBuffer b = new StringBuffer();
		if (expressions == null) return b.toString();

		// print symbol table
		Iterator i = (new TreeSet(environment.symbols.keySet())).iterator();
		b.append("/* Symbol Table */\n");
		while (i.hasNext()) {
			String k = (String)i.next();
			b.append ("// " + k + " = ");
			Symbol s = (Symbol)environment.symbols.get(k);
			if (s != null) b.append(s.toString());
			b.append("\n");
		}
		b.append("\n");

		// print program code
		b.append("/* Program Code */\n");
		for (int j=0; j<listing.length; ++j) {
			b.append(listing[j]);
			b.append("\n");
		}

		return b.toString();
	}

	public void toVerilog(String fn)
	throws IOException {
		FileWriter fw = new FileWriter(new File(fn + ".v"));
		Formatter1 fm = new Formatter1();
		fw.write(fm.formatVerilog(fn, program));
		fw.close();
	}
}
